#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
@version: ??
@author : renzhg
@license: Geneplus 
@contact: renzhg@geneplus.org.cn
@file   : __init__.py.py
@time   : 2024/2/3  21:11
@readme :??
"""


class ClassName(object):

    def __init__(self):
        pass

    def main(self):
        pass

    def run(self):
        self.main()


def func():
    class_name = ClassName()
    class_name.run()
    return


if __name__ == '__main__':
    func()
